/*

        @\^񃂃j^

        SmartMonitor
            DEVICE: ATtiny85V
            CLOCK:  Internal RC OSC. 8MHz / CKDIV = OFF

        Copyrights 2010, Yoshi Hiroshi.

 */

#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdlib.h>

#define REF_VOLT                    1100L
#define LIMIT_MONITOR_VOLT          (1800L << 4)
#define LIMIT_LOW_VOLT              (2100L << 4)
#define LIMIT_MIDDLE_VOLT           (2250L << 4)
#define LIMIT_FULL_VOLT             (2380L << 4)
#define LIMIT_WARNING_VOLT          (2420L << 4)
#define LIMIT_TARGET_VOLT           (2400L << 4)

#define CONV_CAP_VOLT(v)            (((long)(v) * 168L * (REF_VOLT / 64)) / (68L * (1024L / 64)))
#define CONV_REF_VOLT(v)            (((long)(v) * (100L) * (REF_VOLT / 64)) / (36L * (1024L / 64)))

#define BYPASS_PORT_ON              PORTB |= 0x02
#define BYPASS_PORT_OFF             PORTB &= ~0x02
#define BYPASS_ON                   setPwm(0);BYPASS_PORT_ON
#define BYPASS_OFF                  setPwm(0);BYPASS_PORT_OFF
#define BYPASS_PWM(d)               setPwm(d)
#define MAINT_REQ_ON                {PORTB |= 0x04;}
#define MAINT_REQ_OFF               {PORTB &= ~0x04;}
#define CHARGE_END_ON               {PORTB |= 0x01;}
#define CHARGE_END_OFF              {PORTB &= ~0x01;}

#define NUM_HISTORY                  8  // most recent value
#define NUM_HISTORY_S               31  // recent 30 sec
volatile long capV[NUM_HISTORY];        // 16{l
volatile long refV[NUM_HISTORY];        // 16{l
volatile long capVs[NUM_HISTORY_S];    // 128{l
volatile long refVs[NUM_HISTORY_S];    // 128{l
volatile unsigned char capVindex;
volatile unsigned char capVindexS;
volatile long capVolt;       // 16{l
volatile long refVolt;       // 16{l
volatile unsigned char refVoltError;

volatile long capRemain;     //
volatile long refRemain;     //
volatile long capAvrg;      // 128{l
volatile long refAvrg;      // 128{l
volatile long capAvrgS;     // 128{l
volatile long refAvrgS;     // 128{l
volatile long capTime;
volatile long refTime;

#define PWM_MODE_SET                pwmMode = SLEEP_MODE_IDLE
#define PWM_MODE_CLEAR              pwmMode = SLEEP_MODE_PWR_DOWN
#define IS_PWM_MODE                 (pwmMode == SLEEP_MODE_IDLE)
#define IS_NOT_PWM_MODE             (pwmMode == SLEEP_MODE_PWR_DOWN)
#define PWM_COUNTER_MAX             50
volatile char pwmMode;
volatile char pwmCycleCount;
volatile char pwmCounter;
volatile char pwmValue;

void adcEnable()
{
    ADCSRA = _BV(ADEN) | _BV(ADPS2) | _BV(ADPS0);
    ADMUX = _BV(REFS1);
}

inline void adcDisable()
{
    ADMUX = 0;
    ADCSRA = 0;
}

void wdtEnable()
{
    MCUSR &= ~_BV(WDRF);
    WDTCR |= _BV(WDCE) | _BV(WDE);
    WDTCR = _BV(WDIE) | _BV(WDP1) | _BV(WDP0);     // 125mS;
}

void saveCapVolt()
{
    capV[capVindex] = capVolt;
    refV[capVindex] = refVolt;
    capVindex++;
    if(capVindex >= NUM_HISTORY) {
        capVindex = 0;
        for(unsigned char index = 0; index < NUM_HISTORY; index++)
        {
            capVs[capVindexS] += capV[index];
            refVs[capVindexS] += refV[index];
        }
        //capVs[capVindexS] /= NUM_HISTORY;
        //refVs[capVindexS] /= NUM_HISTORY;
        capVindexS++;
        if(capVindexS >= NUM_HISTORY_S) {
            capVindexS = 0;
        }
        capVs[capVindexS] = 0;
        refVs[capVindexS] = 0;
    }
}

void checkMaintenance()
{
    if(capVolt > LIMIT_MONITOR_VOLT) {              // LpV^̓d1800mVȏ̎A
        if(refVolt < 10000) {                        // dCڑ
            MAINT_REQ_ON;
            refVoltError = 1;
        } else {
            if(( (capVolt - refVolt) > (refVolt >> 2)) ||
                 ((refVolt - capVolt) > (refVolt >> 2))) {  // 傫(ref41ȏ)ƁAeiXvON 
                MAINT_REQ_ON;
            } else {
                MAINT_REQ_OFF;
            }
            refVoltError = 0;
        }
    } else {
        MAINT_REQ_OFF;
    }
}

unsigned char getPrevIndex(unsigned char inx, unsigned char max)
{
    inx = inx - 1 + max;
    if(inx >= max) inx -= max;
    return inx;
}

unsigned char checkFewSec(unsigned char amount, unsigned char size, int dead, int ratio, long cap, long ref)
{
    if((cap <= capAvrgS + dead) && (ref <= refAvrgS)) return amount;

    capAvrg = 0;
    refAvrg = 0;
    int incCount = 0;

    unsigned char dataCount = 0;
    if(size >= NUM_HISTORY_S) size = NUM_HISTORY_S - 1;
    unsigned char valIndex = capVindexS + NUM_HISTORY_S - size;
    if(valIndex > NUM_HISTORY_S) valIndex -= NUM_HISTORY_S;
    for(unsigned char index = 0; index < size; index++)
    {
        if(valIndex != capVindexS) {
            if(capVs[valIndex] != 0) {
                if(cap > (capVs[valIndex] + dead)) incCount++;
                capAvrg += capVs[valIndex];
                refAvrg += refVs[valIndex];
                dataCount++;
                valIndex++;
                if(valIndex >= NUM_HISTORY_S) valIndex = 0;
            }
        }
    }
    if(dataCount > (size * 3 / 4)) {
        capAvrg /= dataCount;
        refAvrg /= dataCount;
        long refDelta = ref - refAvrgS;
        long capDelta = cap - capAvrgS;
        if(refDelta <= 0) refDelta = 1;
        if(capDelta <= 0) capDelta = 1;
        if((cap > capAvrg) && (ref > refAvrg)) {     // d㏸
            refTime = (refRemain << 10) / refDelta;
            capTime = (capRemain << 10) / capDelta;
            if(refTime > capTime) {
                unsigned char amountNew = ratio * (refTime - capTime) / capTime ;
                if(amountNew > ratio) amountNew = ratio;
                if(amount < amountNew)
                    amount = amountNew;
            }
        }
    }
    return amount;
}

unsigned char inCharge()
{
    unsigned char amount = 0;

    if(capVolt < (refVolt + 16)) return amount;
    unsigned char prevIndex = getPrevIndex(capVindexS, NUM_HISTORY_S);
    if(((capVolt) << 3) <= capVs[prevIndex]) return amount;
    if(((refVolt) << 3) <= refVs[prevIndex]) return amount;

    capRemain = LIMIT_TARGET_VOLT - capVolt;
    refRemain = LIMIT_TARGET_VOLT - refVolt;

    capAvrgS = 0;
    refAvrgS = 0;
    for(unsigned char index = 0; index < NUM_HISTORY_S; index++)
    {
        if(index != capVindexS) {
            capAvrgS += capVs[index];
            refAvrgS += refVs[index];
        }
    }
    capAvrgS /= (NUM_HISTORY_S - 1);
    refAvrgS /= (NUM_HISTORY_S - 1);

    if(refVoltError == 0) {

// ߂1bf[^~30
        amount = checkFewSec(amount, 30,  16,  10, (capVolt << 3), (refVolt << 3));
// ߂1bf[^~10
        amount = checkFewSec(amount, 10, 160,  40, (capVolt << 3), (refVolt << 3));
        amount = checkFewSec(amount, 10, 512, 100, (capVolt << 3), (refVolt << 3));

    } else {
        if(capVolt > LIMIT_MIDDLE_VOLT)
            amount = 100 * (capVolt - LIMIT_MIDDLE_VOLT) / 
                     (LIMIT_TARGET_VOLT - LIMIT_MIDDLE_VOLT);
    }

    return amount;
}

void setPwm(int val)            // valF%wBől100
{
    val /= 2;
    if(val == 0) {
        pwmValue = 0;
        PWM_MODE_CLEAR;
    } else if(val < PWM_COUNTER_MAX) {
        pwmValue = val;
        PWM_MODE_SET;
    } else {
        pwmValue = PWM_COUNTER_MAX;
        PWM_MODE_CLEAR;
    }
}

inline void bypassCurrent()
{
    unsigned char amountOfBypass;
    if(capVolt >= LIMIT_WARNING_VOLT) {
        CHARGE_END_ON;
        BYPASS_ON;
    } else if(capVolt < LIMIT_LOW_VOLT) {
        CHARGE_END_OFF;
        BYPASS_OFF;
    } else {
        amountOfBypass = inCharge();
        if(amountOfBypass != 0) {
            BYPASS_PWM(amountOfBypass);
            if(capVolt < LIMIT_FULL_VOLT) {
                CHARGE_END_OFF;
            } else {
                CHARGE_END_ON;
            }
        } else {
            BYPASS_OFF;
            if(capVolt < LIMIT_FULL_VOLT) {
                CHARGE_END_OFF;
            } else {
                CHARGE_END_ON;
            }
        }

    }
}

void getAdcData()
{
    adcEnable();            // ADCN
    _delay_us(100);
    ADMUX = _BV(REFS1) | 0x02;
    ADCSRA |= _BV(ADSC);
    loop_until_bit_is_set(ADCSRA, ADIF);    // Ignore First Data
    ADCSRA |= _BV(ADIF);
    refVolt = 0;
    capVolt = 0;
    for(char cnt = 0; cnt < 16; cnt++)      // f[^̎擾(16f[^̕ϒlj
    {
        ADMUX = _BV(REFS1) | 0x02;          // [dd擾
        ADCSRA |= _BV(ADSC);
        loop_until_bit_is_set(ADCSRA, ADIF);
        ADCSRA |= _BV(ADIF);
        refVolt += ADC;
    }

    for(char cnt = 0; cnt < 16; cnt++)      // f[^̎擾(16f[^̕ϒlj
    {
        ADMUX = _BV(REFS1) | 0x03;          // LpV^d擾
        ADCSRA |= _BV(ADSC);
        loop_until_bit_is_set(ADCSRA, ADIF);
        ADCSRA |= _BV(ADIF);
        capVolt += ADC;                          //
    }
    adcDisable();           // ȓd͂̂ADC~

    //refVolt = refVolt >> 4;
    refVolt = CONV_REF_VOLT(refVolt);
    //capVolt = capVolt >> 4;
    capVolt = CONV_CAP_VOLT(capVolt);
}

ISR(TIM0_COMPA_vect)
{
    if(pwmValue == 0) {
        BYPASS_PORT_OFF;
    } else if(pwmValue == PWM_COUNTER_MAX) {
        BYPASS_PORT_ON;
    } else {
        pwmCounter++;
        if(pwmCounter > pwmValue) {
            BYPASS_PORT_OFF;
        } else {
            BYPASS_PORT_ON;
        }
    }
    if(pwmCounter >= PWM_COUNTER_MAX) {
        if(pwmValue != PWM_COUNTER_MAX) {
            BYPASS_PORT_OFF;
        } else {
            BYPASS_PORT_ON;
        }
        pwmCycleCount++;
        pwmCounter = 0;
    }
}

ISR(WDT_vect)
{
}

int main()
{
    CLKPR = _BV(CLKPCE);
    CLKPR = 0x00;
    BYPASS_OFF;
    MAINT_REQ_OFF;
    CHARGE_END_OFF;
    DDRB |= 0x07;
    DIDR0 = 0x3C;
    capVindex = 0;
    capVindexS = 0;
    capVolt = 0;
    refVolt = 0;
    refVoltError = 0;
    PWM_MODE_CLEAR;
    OCR0A = 99;     // 0.1mS
    TCCR0A = _BV(WGM01);
    TCCR0B = _BV(CS01);
    wdtEnable();

    while(1)
    {
        set_sleep_mode(pwmMode);
        sleep_enable();
        sei();
        sleep_cpu();
        sleep_disable();

        if(IS_NOT_PWM_MODE || (pwmCycleCount >= 25)) {  // 125mS
            pwmCycleCount = 0;
            //TIMSK &= ~_BV(OCIE0A);

            getAdcData();
            checkMaintenance();
            bypassCurrent();        // dʂɂ[ddoCpX
            saveCapVolt();

            if(IS_PWM_MODE) {
                TIMSK |= _BV(OCIE0A);
            } else {
                TIMSK &= ~_BV(OCIE0A);
            }
        }
    }
}
